@@ -14,7 +14,6 @@ __maintainer__ = 'Anubhav Jain'
 __email__ = 'ajain@lbl.gov'
 __date__ = 'Dec 12, 2012'
 
-
 def singleton(class_):
     instances = {}
 
@@ -25,7 +24,6 @@ def singleton(class_):
 
     return getinstance
 
-
 @singleton
 class FWConfig(object):
     def __init__(self):
@@ -66,10 +64,6 @@ class FWConfig(object):
 
         self.CONFIG_FILE_DIR = '.'  # directory containing config files (if not individually set)
 
-        self.QSTAT_FREQUENCY = 50  # set this higher to avoid qstats, lower to alwas
-
-        self.ALWAYS_CREATE_NEW_BLOCK = False  # always create new block on queue launcher call
-
         self.override_user_settings()
 
     def override_user_settings(self):
@@ -89,9 +83,7 @@ class FWConfig(object):
                         self.USER_PACKAGES.extend(v)
                     elif key == 'ECHO_TEST':
                         print v
-                    elif not hasattr(self, key):
-                        raise ValueError(
-                            'Invalid FW_config file has unknown parameter: {}'.format(
-                                key))
+                    elif not getattr(self, key, None):
+                        raise ValueError('Invalid FW_config file has unknown parameter: {}'.format(key))
                     else:
                         self.__setattr__(key, v)
