@@ -4,19 +4,13 @@
 The LaunchPad manages the FireWorks database.
 """
 import datetime
-import json
-import os
 import time
-import traceback
-
-from pymongo.mongo_client import MongoClient
-from pymongo import DESCENDING
-
 from fireworks.core.fw_config import FWConfig
 from fireworks.utilities.fw_serializers import FWSerializable
-from fireworks.core.firework import FireWork, Launch, Workflow, FWAction, Tracker
-from fireworks.utilities.fw_utilities import get_fw_logger
-
+from pymongo.mongo_client import MongoClient
+from fireworks.core.firework import FireWork, Launch, Workflow
+from pymongo import DESCENDING
+from fireworks.utilities.fw_utilities import get_fw_logger, log_exception
 
 __author__ = 'Anubhav Jain'
 __copyright__ = 'Copyright 2013, The Materials Project'
@@ -42,7 +36,7 @@ class LaunchPad(FWSerializable):
     """
 
     def __init__(self, host='localhost', port=27017, name='fireworks', username=None, password=None,
-                 logdir=None, strm_lvl=None, user_indices=None, wf_user_indices=None):
+                 logdir=None, strm_lvl=None):
         """
         
         :param host:
@@ -52,8 +46,6 @@ class LaunchPad(FWSerializable):
         :param password:
         :param logdir:
         :param strm_lvl:
-        :param user_indices:
-        :param wf_user_indices:
         """
         self.host = host
         self.port = port
@@ -66,47 +58,29 @@ class LaunchPad(FWSerializable):
         self.strm_lvl = strm_lvl if strm_lvl else 'INFO'
         self.m_logger = get_fw_logger('launchpad', l_dir=self.logdir, stream_level=self.strm_lvl)
 
-        self.user_indices = user_indices if user_indices else []
-        self.wf_user_indices = wf_user_indices if wf_user_indices else []
-
-        # get connection
         self.connection = MongoClient(host, port, j=True)
-        self.db = self.connection[name]
+        self.database = self.connection[name]
         if username:
-            self.db.authenticate(username, password)
+            self.database.authenticate(username, password)
 
-        self.fireworks = self.db.fireworks
-        self.launches = self.db.launches
-        self.offline_runs = self.db.offline_runs
-        self.fw_id_assigner = self.db.fw_id_assigner
-        self.workflows = self.db.workflows
+        self.fireworks = self.database.fireworks
+        self.launches = self.database.launches
+        self.fw_id_assigner = self.database.fw_id_assigner
+        self.workflows = self.database.workflows
 
     def to_dict(self):
         """
         Note: usernames/passwords are exported as unencrypted Strings!
         """
         d = {'host': self.host, 'port': self.port, 'name': self.name, 'username': self.username,
-             'password': self.password, 'logdir': self.logdir, 'strm_lvl': self.strm_lvl,
-             'user_indices': self.user_indices, 'wf_user_indices': self.wf_user_indices}
+             'password': self.password, 'logdir': self.logdir, 'strm_lvl': self.strm_lvl}
         return d
 
     @classmethod
     def from_dict(cls, d):
         logdir = d.get('logdir', None)
         strm_lvl = d.get('strm_lvl', None)
-        user_indices = d.get('user_indices', [])
-        wf_user_indices = d.get('wf_user_indices', [])
-        return LaunchPad(d['host'], d['port'], d['name'], d['username'], d['password'], logdir,
-                         strm_lvl, user_indices, wf_user_indices)
-
-    @classmethod
-    def auto_load(cls):
-        if FWConfig().LAUNCHPAD_LOC:
-            return LaunchPad.from_file(FWConfig().LAUNCHPAD_LOC)
-        elif FWConfig().CONFIG_FILE_DIR:
-            return LaunchPad.from_file(os.path.join(FWConfig().CONFIG_FILE_DIR, 'my_launchpad.yaml'))
-        return LaunchPad()
-
+        return LaunchPad(d['host'], d['port'], d['name'], d['username'], d['password'], logdir, strm_lvl)
 
     def reset(self, password, require_password=True):
         """
@@ -124,7 +98,6 @@ class LaunchPad(FWSerializable):
             self.fireworks.remove()
             self.launches.remove()
             self.workflows.remove()
-            self.offline_runs.remove()
             self._restart_ids(1, 1)
             self.tuneup()
             self.m_logger.info('LaunchPad was RESET.')
@@ -138,13 +111,12 @@ class LaunchPad(FWSerializable):
             self.m_logger.info('Performing maintenance on Launchpad...')
 
             self.m_logger.debug('Tracking down FIZZLED jobs...')
-            fl, ff = self.detect_lostruns(fizzle=True)
+            fl = self.detect_fizzled(fix=True)
             if fl:
                 self.m_logger.info('Detected {} FIZZLED launches: {}'.format(len(fl), fl))
-                self.m_logger.info('Detected {} FIZZLED FWs: {}'.format(len(ff), ff))
 
             self.m_logger.debug('Tracking down stuck RESERVED jobs...')
-            ur = self.detect_unreserved(rerun=True)
+            ur = self.detect_unreserved(fix=True)
             if ur:
                 self.m_logger.info('Unreserved {} RESERVED launches: {}'.format(len(ur), ur))
 
@@ -156,7 +128,7 @@ class LaunchPad(FWSerializable):
             self.m_logger.debug('Sleeping for {} secs...'.format(maintain_interval))
             time.sleep(maintain_interval)
 
-    def add_wf(self, wf, reassign_all=True):
+    def add_wf(self, wf):
         """
 
         :param wf: a Workflow object.
@@ -166,7 +138,7 @@ class LaunchPad(FWSerializable):
             wf = Workflow.from_FireWork(wf)
 
         # insert the FireWorks and get back mapping of old to new ids
-        old_new = self._upsert_fws(wf.id_fw.values(), reassign_all=reassign_all)
+        old_new = self._upsert_fws(wf.id_fw.values(), reassign_all=True)
 
         # update the Workflow with the new ids
         wf._reassign_ids(old_new)
@@ -232,45 +204,23 @@ class LaunchPad(FWSerializable):
         fws = []
         for fw_id in links_dict['nodes']:
             fws.append(self.get_fw_by_id(fw_id))
+        links = Workflow.Links.from_dict(links_dict['links']).to_dict()  # necessary because Mongo no like int keys
 
-        return Workflow(fws, links_dict['links'], links_dict['name'], links_dict['metadata'])
+        return Workflow(fws, links, links_dict['metadata'])
 
-    def get_fw_ids(self, query=None, sort=None, limit=0, count_only=False):
+    def get_fw_ids(self, query=None, sort=False):
         """
         Return all the fw ids that match a query,
-        :param query: (dict) representing a Mongo query
-        :param sort: [(str,str)] sort argument in Pymongo format
-        :param limit: (int) limit the results
-        :param count_only: (bool) only return the count rather than explicit ids
+        :param query: a dict representing a Mongo query
         """
         fw_ids = []
         criteria = query if query else {}
-
-        if count_only:
-            return self.fireworks.find(criteria, {"fw_id": True}, sort=sort).limit(limit).count()
-
-        for fw in self.fireworks.find(criteria, {"fw_id": True}, sort=sort).limit(limit):
+        sort = [("spec._priority", DESCENDING)] if sort else None
+        for fw in self.fireworks.find(criteria, {"fw_id": True}, sort=sort):
             fw_ids.append(fw["fw_id"])
+
         return fw_ids
 
-    def get_wf_ids(self, query=None, sort=None, limit=0, count_only=False):
-        """
-        Return one fw id for all workflows that match a query,
-        :param query: (dict) representing a Mongo query
-        :param sort: [(str,str)] sort argument in Pymongo format
-        :param limit: (int) limit the results
-        :param count_only: (bool) only return the count rather than explicit ids
-        """
-        wf_ids = []
-        criteria = query if query else {}
-        if count_only:
-            return self.workflows.find(criteria, {"nodes": True}, sort=sort).limit(limit).count()
-
-        for fw in self.workflows.find(criteria, {"nodes": True}, sort=sort).limit(limit):
-            wf_ids.append(fw["nodes"][0])
-
-        return wf_ids
-
     def run_exists(self, fworker=None):
         """
         Checks to see if the database contains any FireWorks that are ready to run
@@ -286,47 +236,28 @@ class LaunchPad(FWSerializable):
         self.fireworks.ensure_index('fw_id', unique=True)
         self.fireworks.ensure_index('state')
         self.fireworks.ensure_index('spec._category')
-        self.fireworks.ensure_index('created_on')
-        self.fireworks.ensure_index('name')
-
         self.launches.ensure_index('launch_id', unique=True)
         self.launches.ensure_index('state')
         self.launches.ensure_index('time_start')
         self.launches.ensure_index('time_end')
         self.launches.ensure_index('host')
         self.launches.ensure_index('ip')
-        self.launches.ensure_index('fworker.name')
-
-        self.workflows.ensure_index('name')
-        self.workflows.ensure_index('created_on')
-        self.workflows.ensure_index('updated_on')
-        self.workflows.ensure_index('nodes')
-
-        for idx in self.user_indices:
-            self.fireworks.ensure_index(idx)
-
-        for idx in self.wf_user_indices:
-            self.workflows.ensure_index(idx)
 
         self.m_logger.debug('Compacting database...')
         try:
-            self.db.command({'compact': 'fireworks'})
-            self.db.command({'compact': 'launches'})
+            self.database.command({'compact': 'fireworks'})
+            self.database.command({'compact': 'launches'})
         except:
             self.m_logger.debug('Database compaction failed (not critical)')
 
 
     def defuse_fw(self, fw_id):
         allowed_states = ['DEFUSED', 'WAITING', 'READY', 'FIZZLED']
-        f = self.fireworks.find_and_modify({'fw_id': fw_id, 'state': {'$in': allowed_states}},
+        return self.fireworks.find_and_modify({'fw_id': fw_id, 'state': {'$in': allowed_states}},
                                               {'$set': {'state': 'DEFUSED'}})
 
-        self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
-        return f
-
     def reignite_fw(self, fw_id):
-        f = self.fireworks.find_and_modify({'fw_id': fw_id, 'state': 'DEFUSED'},
-                                           {'$set': {'state': 'WAITING'}})
+        f = self.fireworks.find_and_modify({'fw_id': fw_id, 'state': 'DEFUSED'}, {'$set': {'state': 'WAITING'}})
         if f:
             self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
         return f
@@ -336,28 +267,11 @@ class LaunchPad(FWSerializable):
         for fw in wf.fws:
             self.defuse_fw(fw.fw_id)
 
-        self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
-
-
     def reignite_wf(self, fw_id):
         wf = self.get_wf_by_fw_id(fw_id)
         for fw in wf.fws:
             self.reignite_fw(fw.fw_id)
 
-    def archive_wf(self, fw_id):
-        # first archive all the launches, so they are not used in duplicate checks
-        wf = self.get_wf_by_fw_id(fw_id)
-        fw_ids = [f.fw_id for f in wf.fws]
-        for fw_id in fw_ids:
-            self.rerun_fw(fw_id)
-
-        # second set the state of all FWs to ARCHIVED
-        wf = self.get_wf_by_fw_id(fw_id)
-        for fw in wf.fws:
-            self.fireworks.find_and_modify({'fw_id': fw.fw_id}, {'$set': {'state': 'ARCHIVED'}})
-
-        self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
-
     def _restart_ids(self, next_fw_id, next_launch_id):
         """
         (internal method) Used to reset id counters
@@ -366,11 +280,9 @@ class LaunchPad(FWSerializable):
         :param next_launch_id: id to give next Launch (int)
         """
         self.fw_id_assigner.remove()
-        self.fw_id_assigner.find_and_modify({'_id': -1}, {'next_fw_id': next_fw_id,
-                                                          'next_launch_id': next_launch_id},
+        self.fw_id_assigner.find_and_modify({'_id': -1}, {'next_fw_id': next_fw_id, 'next_launch_id': next_launch_id},
                                             upsert=True)
-        self.m_logger.debug(
-            'RESTARTED fw_id, launch_id to ({}, {})'.format(next_fw_id, next_launch_id))
+        self.m_logger.debug('RESTARTED fw_id, launch_id to ({}, {})'.format(next_fw_id, next_launch_id))
 
     def _check_fw_for_uniqueness(self, m_fw):
         # check if there are duplicates
@@ -400,13 +312,12 @@ class LaunchPad(FWSerializable):
                 m_fw = self.fireworks.find_and_modify(m_query, {'$set': {'state': 'RESERVED'}},
                                                       sort=[("spec._priority", DESCENDING)])
             else:
-                m_fw = self.fireworks.find_one(m_query, {'fw_id': 1, 'spec': 1},
-                                               sort=[("spec._priority", DESCENDING)])
-                
+                m_fw = self.fireworks.find_one(m_query, {'fw_id': 1}, sort=[("spec._priority", DESCENDING)])
+
             if not m_fw:
                 return None
-
             m_fw = self.get_fw_by_id(m_fw['fw_id'])
+
             if self._check_fw_for_uniqueness(m_fw):
                 return m_fw
 
@@ -417,9 +328,7 @@ class LaunchPad(FWSerializable):
             # create a launch
         # TODO: this code is duplicated with checkout_fw with minimal mods, should refactor this!!
         launch_id = self.get_new_launch_id()
-        trackers = [Tracker.from_dict(f) for f in m_fw.spec['_trackers']] if '_trackers' in m_fw.spec else None
-        m_launch = Launch('RESERVED', launch_dir, fworker, host, ip, trackers=trackers, launch_id=launch_id,
-                          fw_id=m_fw.fw_id)
+        m_launch = Launch('RESERVED', launch_dir, fworker, host, ip, launch_id=launch_id, fw_id=m_fw.fw_id)
         self._upsert_launch(m_launch)
 
         # add launch to FW
@@ -439,22 +348,21 @@ class LaunchPad(FWSerializable):
         for fw in self.fireworks.find({'launches': launch_id, 'state': 'RESERVED'}, {'fw_id': 1}):
             self.fireworks.find_and_modify({'fw_id': fw['fw_id']}, {'$set': {'state': 'READY'}})
 
-    def detect_unreserved(self, expiration_secs=FWConfig().RESERVATION_EXPIRATION_SECS, rerun=False):
+    def detect_unreserved(self, expiration_secs=FWConfig().RESERVATION_EXPIRATION_SECS, fix=False):
         bad_launch_ids = []
         now_time = datetime.datetime.utcnow()
         cutoff_timestr = (now_time - datetime.timedelta(seconds=expiration_secs)).isoformat()
         bad_launch_data = self.launches.find({'state': 'RESERVED', 'state_history': {
-            '$elemMatch': {'state': 'RESERVED', 'updated_on': {'$lte': cutoff_timestr}}}},
-                                             {'launch_id': 1})
+            '$elemMatch': {'state': 'RESERVED', 'updated_on': {'$lte': cutoff_timestr}}}}, {'launch_id': 1})
         for ld in bad_launch_data:
             bad_launch_ids.append(ld['launch_id'])
-        if rerun:
+        if fix:
             for lid in bad_launch_ids:
                 self.unreserve(lid)
         return bad_launch_ids
 
     def mark_fizzled(self, launch_id):
-        # TODO: this seems a lot like the code in complete_launch...DRY
+        # TODO: this seems a lot like the code in _complete_launch...DRY
 
         # Do a confirmed write and make sure state_history is preserved
         m_launch = self.get_launch_by_id(launch_id)
@@ -466,44 +374,25 @@ class LaunchPad(FWSerializable):
             wf = self.get_wf_by_fw_id(fw_id)
             self._refresh_wf(wf, fw_id)
 
-    def detect_lostruns(self, expiration_secs=FWConfig().RUN_EXPIRATION_SECS, fizzle=False, rerun=False, max_runtime=None):
+    def detect_fizzled(self, expiration_secs=FWConfig().RUN_EXPIRATION_SECS, fix=False):
         bad_launch_ids = []
-        bad_fw_ids = []
         now_time = datetime.datetime.utcnow()
         cutoff_timestr = (now_time - datetime.timedelta(seconds=expiration_secs)).isoformat()
         bad_launch_data = self.launches.find({'state': 'RUNNING', 'state_history': {
-            '$elemMatch': {'state': 'RUNNING', 'updated_on': {'$lte': cutoff_timestr}}}},
-                                             {'launch_id': 1, 'fw_id': 1})
+            '$elemMatch': {'state': 'RUNNING', 'updated_on': {'$lte': cutoff_timestr}}}}, {'launch_id': 1})
         for ld in bad_launch_data:
-            bad_launch = False
-            if max_runtime:
-                m_l = self.get_launch_by_id(ld['launch_id'])
-                utime = m_l._get_time('RUNNING', use_update_time=True)
-                ctime = m_l._get_time('RUNNING', use_update_time=False)
-                if (utime-ctime).seconds <= max_runtime:
-                    bad_launch=True
-
-            if not max_runtime or bad_launch:
-                bad_launch_ids.append(ld['launch_id'])
-                bad_fw_ids.append(ld['fw_id'])
-        if fizzle or rerun:
+            bad_launch_ids.append(ld['launch_id'])
+        if fix:
             for lid in bad_launch_ids:
                 self.mark_fizzled(lid)
-        if rerun:
-            for fw_id in bad_fw_ids:
-                # only rerun if the FireWork has 'FIZZLED' state. It is possible that
-                # a different Launch for the same FireWork is 'COMPLETED'
-                if self.fireworks.find_one({"fw_id": fw_id, "state": "FIZZLED"}):
-                    self.rerun_fw(fw_id)
-
-        return bad_launch_ids, bad_fw_ids
+        return bad_launch_ids
 
     def _set_reservation_id(self, launch_id, reservation_id):
         m_launch = self.get_launch_by_id(launch_id)
         m_launch.set_reservation_id(reservation_id)
         self.launches.find_and_modify({'launch_id': launch_id}, m_launch.to_db_dict())
 
-    def checkout_fw(self, fworker, launch_dir, fw_id=None, host=None, ip=None):
+    def _checkout_fw(self, fworker, launch_dir, fw_id=None, host=None, ip=None):
         """
         (internal method) Finds a FireWork that's ready to be run, marks it as running,
         and returns it to the caller. The caller is responsible for running the FireWork.
@@ -528,9 +417,7 @@ class LaunchPad(FWSerializable):
 
         state_history = reserved_launch.state_history if reserved_launch else None
         l_id = reserved_launch.launch_id if reserved_launch else self.get_new_launch_id()
-        trackers = [Tracker.from_dict(f) for f in m_fw.spec['_trackers']] if '_trackers' in m_fw.spec else None
-        m_launch = Launch('RUNNING', launch_dir, fworker, host, ip, trackers=trackers, state_history=state_history,
-                          launch_id=l_id,
+        m_launch = Launch('RUNNING', launch_dir, fworker, host, ip, state_history=state_history, launch_id=l_id,
                           fw_id=m_fw.fw_id)
 
         self._upsert_launch(m_launch)
@@ -542,16 +429,14 @@ class LaunchPad(FWSerializable):
             m_fw.launches.append(m_launch)
         else:
             # we're updating an existing launch
-            m_fw.launches = [m_launch if l.launch_id == m_launch.launch_id else l for l in
-                             m_fw.launches]
+            m_fw.launches = [m_launch if l.launch_id == m_launch.launch_id else l for l in m_fw.launches]
 
         m_fw.state = 'RUNNING'
         self._upsert_fws([m_fw])
 
         # update any duplicated runs
-        for fw in self.fireworks.find(
-                {'launches': l_id, 'state': {'$in': ['WAITING', 'READY', 'RESERVED', 'FIZZLED']}},
-                {'fw_id': 1}):
+        for fw in self.fireworks.find({'launches': l_id, 'state': {'$in': ['WAITING', 'READY', 'RESERVED', 'FIZZLED']}},
+                                      {'fw_id': 1}):
             fw_id = fw['fw_id']
             fw = self.get_fw_by_id(fw_id)
             fw.state = 'RUNNING'
@@ -561,12 +446,7 @@ class LaunchPad(FWSerializable):
 
         return m_fw, l_id
 
-    def _change_launch_dir(self, launch_id, launch_dir):
-        m_launch = self.get_launch_by_id(launch_id)
-        m_launch.launch_dir = launch_dir
-        self._upsert_launch(m_launch)
-
-    def complete_launch(self, launch_id, action, state='COMPLETED'):
+    def _complete_launch(self, launch_id, action, state='COMPLETED'):
         """
         (internal method) used to mark a FireWork's Launch as completed.
         :param launch_id:
@@ -582,18 +462,11 @@ class LaunchPad(FWSerializable):
         for fw in self.fireworks.find({'launches': launch_id}, {'fw_id': 1}):
             fw_id = fw['fw_id']
             self._refresh_wf(self.get_wf_by_fw_id(fw_id), fw_id)
-        # change return type to dict to make return type seriazlizable to
-        # support job packing
-        return m_launch.to_dict()
 
-    def ping_launch(self, launch_id, ptime=None):
+    def _ping_launch(self, launch_id):
         m_launch = self.get_launch_by_id(launch_id)
-        for tracker in m_launch.trackers:
-            tracker.track_file(m_launch.launch_dir)
-
-        m_launch.touch_history(ptime)
-        self.launches.update({'launch_id': launch_id, 'state': 'RUNNING'},
-            {'$set':{'state_history':m_launch.to_db_dict()['state_history'], 'trackers': [t.to_dict() for t in m_launch.trackers]}})
+        m_launch.touch_history()
+        self.launches.find_and_modify({'launch_id': launch_id}, m_launch.to_db_dict())
 
     def get_new_fw_id(self):
         """
@@ -605,8 +478,7 @@ class LaunchPad(FWSerializable):
         """
         Checkout the next Launch id
         """
-        return self.fw_id_assigner.find_and_modify({}, {'$inc': {'next_launch_id': 1}})[
-            'next_launch_id']
+        return self.fw_id_assigner.find_and_modify({}, {'$inc': {'next_launch_id': 1}})['next_launch_id']
 
     def _upsert_fws(self, fws, reassign_all=False):
         old_new = {} # mapping between old and new FireWork ids
@@ -646,9 +518,8 @@ class LaunchPad(FWSerializable):
         updated_fws = [wf.id_fw[fid] for fid in updated_ids]
         old_new = self._upsert_fws(updated_fws)
         wf._reassign_ids(old_new)
-        # redo the links - note that if you don't search across all keys, you can get errors
-        self.workflows.find_and_modify({'nodes': {'$in': wf.id_fw.keys()}}, wf.to_db_dict())
-
+        # redo the links
+        self.workflows.find_and_modify({'nodes': wf.fws[0].fw_id}, wf.to_db_dict())
 
     def _steal_launches(self, thief_fw):
         stolen = False
@@ -661,96 +532,43 @@ class LaunchPad(FWSerializable):
             self.m_logger.debug('Querying for duplicates, fw_id: {}'.format(thief_fw.fw_id))
             for potential_match in self.fireworks.find(m_query):
                 self.m_logger.debug(
-                    'Verifying for duplicates, fw_ids: {}, {}'.format(thief_fw.fw_id,
-                                                                      potential_match['fw_id']))
+                    'Verifying for duplicates, fw_ids: {}, {}'.format(thief_fw.fw_id, potential_match['fw_id']))
                 spec1 = dict(thief_fw.to_dict()['spec'])  # defensive copy
                 spec2 = dict(potential_match['spec'])  # defensive copy
                 if m_dupefinder.verify(spec1, spec2):  # verify the match
                     # steal the launches
                     victim_fw = self.get_fw_by_id(potential_match['fw_id'])
                     thief_launches = [l.launch_id for l in thief_fw.launches]
-                    valuable_launches = [l for l in victim_fw.launches if
-                                         l.launch_id not in thief_launches]
+                    valuable_launches = [l for l in victim_fw.launches if l.launch_id not in thief_launches]
                     for launch in valuable_launches:
                         thief_fw.launches.append(launch)
                         stolen = True
                         self.m_logger.info(
-                            'Duplicate found! fwids {} and {}'.format(thief_fw.fw_id,
-                                                                      potential_match['fw_id']))
+                            'Duplicate found! fwids {} and {}'.format(thief_fw.fw_id, potential_match['fw_id']))
         return stolen
 
     def _upsert_launch(self, m_launch):
-        # TODO: this no longer needs to be its own function (much was removed)
+        # Do a confirmed write of Launch
+
         self.launches.find_and_modify({'launch_id': m_launch.launch_id}, m_launch.to_db_dict(), upsert=True)
 
-    def set_priority(self, fw_id, priority):
-        self.fireworks.find_and_modify({"fw_id": fw_id}, {'$set': {'spec._priority': priority}})
-
-    def get_logdir(self):
-        # AJ: This is needed for job packing due to Proxy objects not being fully featured...
-        return self.logdir
-
-    def add_offline_run(self, launch_id, fw_id, name):
-        d = {'fw_id': fw_id}
-        d['launch_id'] = launch_id
-        d['name'] = name
-        d['created_on'] = datetime.datetime.utcnow().isoformat()
-        d['updated_on'] = datetime.datetime.utcnow().isoformat()
-        d['deprecated'] = False
-        d['completed'] = False
-        self.offline_runs.insert(d)
-
-    def recover_offline(self, launch_id, ignore_errors=False):
-        # get the launch directory
-        m_launch = self.get_launch_by_id(launch_id)
-        try:
-            self.m_logger.debug("RECOVERING fw_id: {}".format(m_launch.fw_id))
-            # look for ping file - update the FireWork if this is the case
-            ping_loc = os.path.join(m_launch.launch_dir, "FW_ping.json")
-            if os.path.exists(ping_loc):
-                with open(ping_loc) as f:
-                    ping_time = datetime.datetime.strptime(json.loads(f.read())['ping_time'], "%Y-%m-%dT%H:%M:%S.%f")
-                    self.ping_launch(launch_id, ping_time)
-
-            # look for action in FW_offline.json
-            offline_loc = os.path.join(m_launch.launch_dir, "FW_offline.json")
-            with open(offline_loc) as f:
-                offline_data = json.loads(f.read())
-                if 'started_on' in offline_data:
-                    m_launch.state = 'RUNNING'
-                    for s in m_launch.state_history:
-                        if s['state'] == 'RUNNING':
-                            s['created_on'] = datetime.datetime.strptime(offline_data['started_on'], "%Y-%m-%dT%H:%M:%S.%f")
-                    self._upsert_launch(m_launch)
-
-                if 'fwaction' in offline_data:
-                    fwaction = FWAction.from_dict(offline_data['fwaction'])
-                    state = offline_data['state']
-                    m_launch = Launch.from_dict(
-                        self.complete_launch(launch_id, fwaction, state))
-                    for s in m_launch.state_history:
-                        if s['state'] == offline_data['state']:
-                            s['created_on'] = datetime.datetime.strptime(offline_data['completed_on'], "%Y-%m-%dT%H:%M:%S.%f")
-                    self._upsert_launch(m_launch)
-                    self.offline_runs.update({"launch_id": launch_id}, {"$set": {"completed":True}})
-
-            # update the updated_on
-            self.offline_runs.update({"launch_id": launch_id}, {"$set": {"updated_on": datetime.datetime.utcnow().isoformat()}})
-            return None
-        except:
-            if not ignore_errors:
-                traceback.print_exc()
-            return m_launch.fw_id
-
-    def forget_offline(self, fw_id):
-        self.offline_runs.update({"fw_id": fw_id}, {"$set": {"deprecated":True}})
-
-    def get_tracker_data(self, fw_id):
-        data = []
-        for l in self.launches.find({'fw_id': fw_id}, {'trackers': 1, 'launch_id': 1}):
-            if 'trackers' in l:  # backwards compatibility
-                trackers = [Tracker.from_dict(t) for t in l['trackers']]
-                data.append({'launch_id': l['launch_id'], 'trackers': trackers})
-
-        return data
-
+        l_id = m_launch.launch_id
+        # confirm write
+        # I can't believe this is actually necessary (and yes, it appears to be necessary)
+        nloops = 0
+        while not self.launches.find_one({'launch_id': l_id, 'state': m_launch.state}):
+            self.m_logger.debug('Waiting for a delayed write of Launch...')
+            nloops += 1
+            if nloops == 80:
+                if m_launch.state == 'FIZZLED':
+                    # We're unable to mark the launch as FIZZLED
+                    self.m_logger.critical('UNABLE to mark launch_id: {} as fizzled!'.format(l_id))
+                    break
+                    # mark the launch as FIZZLED because we're unable to update its state...
+                self.m_logger.error('FIZZLED launch id: {} because could not confirm write!!'.format(l_id))
+                self.mark_fizzled(l_id)
+                break
+            if nloops % 20 == 0:
+                self.m_logger.info('Fixing a lost write of Launch!!')
+                self.launches.find_and_modify({'launch_id': l_id}, m_launch.to_db_dict())
+            time.sleep(4)
\ No newline at end of file
